/*
 * Userspace program that communicates with the vga_ball device driver
 * through ioctls
 *
 * Stephen A. Edwards
 * Columbia University
 *
 * scokets ref: https://www.geeksforgeeks.org/socket-programming-cc/
 */

#include <stdio.h>
#include "vga_ball.h"
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include<netinet/in.h>

#include<stdlib.h>
#include<string.h>

#define SERVER_HOST "127.0.0.1"
#define SERVER_PORT 42000

// special address
// 1000_0000_0000_0000
// FPGA interprets writedata as 
// MIDI note values when data 
// is written to this address
#define MIDI_WRITEADDR 0x8000


// special address for env data
// 0100_0000_0000_0000
#define ENV_WRITEADDR 0x8800

#define NSAMPLES 32768 
signed int wav_table[NSAMPLES];

//char *sin_file= "sin_96.txt";
//char *sin_file= "500_zeros.txt";

int vga_ball_fd;

#define MAX_LINESIZE 15

int line_num = 0;
char line [MAX_LINESIZE+1];
char * next_field;

char *ptr ;

int i = 0;

int load_wave_table(char * fn){
    FILE * f = fopen (fn, "r");
    while (1){
        fgets(line, MAX_LINESIZE, f);

        if(feof(f))
            break;

        line_num++;
        next_field = strtok(line, " \n");
        while (next_field != NULL)
        {
            short v = (int)(strtol(next_field, &ptr, 10) * 0.5);
            wav_table[i++] = v;
            //        printf("Line %d:%s \t int:%d\n", line_num, next_field, v);
            next_field = strtok(NULL, " \n");
        }
    }
    fclose(f);
    return 0;
}



/* Read and print the background color */
void print_background_color(const vga_ball_color_t *c) {
    vga_ball_arg_t vla;
    vla.background = *c; 
    if (ioctl(vga_ball_fd, VGA_BALL_READ_BACKGROUND, &vla)) {
        perror("ioctl(VGA_BALL_READ_BACKGROUND) failed");
        return;
    }
    //vla.background = *c; 
    printf("%d %02x \n",
            vla.background.sin, vla.background.loc);
}

/* Set the background color */
void set_background_color(const vga_ball_color_t *c)
{
    vga_ball_arg_t vla;
    vla.background = *c;
    if (ioctl(vga_ball_fd, VGA_BALL_WRITE_BACKGROUND, &vla)) {
        perror("ioctl(VGA_BALL_SET_BACKGROUND) failed");
        return;
    }
}


void write_sample_data(){
    vga_ball_color_t c;
   
/*
       for (int i =0; i < NSAMPLES; i++)
       {
       c.sin[i] = wav_table[i]; // 0xE: 27 & 26 bits of data
    //printf("%d\n", c.sin);
    c.loc = i;
    printf("Writing to addr(dec)\t%d\n", i);
    } 
    set_background_color(&c);
*/

    for (int i = 0; i< NSAMPLES; i++){
        c.sin = wav_table[i];
        c.loc = i;
        set_background_color(&c);
    }

    // READ
/*
    for (int i =0; i < NSAMPLES; i++)
    {
        //printf("%d\n", c.sin);
        c.loc = i;
        //printf("%d\n", c.loc);
        print_background_color(&c);
    }
*/
}


void socket_server_loop(){
    printf("\nStarting socket server\n");

    // data to write, and address to write
    vga_ball_color_t c;
    
    int server_fd, new_socket, valread; 
    struct sockaddr_in address; 
    int opt = 1; 
    int addrlen = sizeof(address); 
    char buffer[32] = {0}; 

    if ((server_fd = socket(AF_INET, SOCK_STREAM, 0)) == 0) 
    { 
        perror("socket failed"); 
        exit(EXIT_FAILURE); 
    } 

    if (setsockopt(server_fd, SOL_SOCKET, SO_REUSEADDR | SO_REUSEPORT, 
                &opt, sizeof(opt))) 
    { 
        perror("setsockopt"); 
        exit(EXIT_FAILURE); 
    } 
    address.sin_family = AF_INET; 
    address.sin_addr.s_addr = INADDR_ANY; 
    address.sin_addr.s_addr = inet_addr(SERVER_HOST); 
    address.sin_port = htons( SERVER_PORT ); 


    if (bind(server_fd, (struct sockaddr *)&address,  
                sizeof(address))<0) 
    { 
        perror("bind failed"); 
        exit(EXIT_FAILURE); 
    } 

    if (listen(server_fd, 3) < 0) 
    { 
        perror("listen failure"); 
        exit(EXIT_FAILURE); 
    }

    if ((new_socket = accept(server_fd, (struct sockaddr *)&address,  
                    (socklen_t*)&addrlen))<0) 
    { 
        perror("accept failure"); 
        exit(EXIT_FAILURE); 
    }

    for (;;){
        // Read socket
        valread = read( new_socket , buffer, 32); 
        //printf("%s\n",buffer );

        // packed struct for storing keypress 
        // values in 16 bit wide val 
        typedef union {
            uint16_t val;
            struct{
                unsigned int na: 5;
                unsigned int nb: 5;
                unsigned int nc: 5;
                unsigned int x: 1;
            };
        }packed_un_t;

        packed_un_t u1 = {.val = 0};

        short vals[3] = {0};
        int i=0;
        
        // Extract comma sep values and convert to int
        const char s[2] = ",";
        char *token;

        token = strtok(buffer, s);
        while(token!= NULL){
            int xval =  (int)strtol(token, NULL, 16); // str to int
            //printf("\nExtracted val: %02x\n",xval);
            vals[i++] = (short)xval;
            token = strtok(NULL, s);  // Get next token
        }
        
        u1.na = vals[0];
        u1.nb = vals[1];
        u1.nc = vals[2];
        
        printf (">>>> %06x\n", u1.val);
        
        // load values to be written to FPGA
        c.sin = u1.val;
        // Addr is the secret address, writing to which makes
        // fpga interpret writeaddr values as midi note
        c.loc = MIDI_WRITEADDR;     
        set_background_color(&c);

    } // for(;;) loop

}

int main(int argc, char *argv[])
{
    vga_ball_arg_t vla;
    int i;
    //static const char filename[] = "/dev/vga_ball";
    static const char filename[] = "/dev/audio_codec";

    printf("VGA ball Userspace program started\n");

    if ( (vga_ball_fd = open(filename, O_RDWR)) == -1) {
        fprintf(stderr, "could not open %s\n", filename);
        return -1;
    }

    printf("initial state: not priting ");

    // No args passed: receive-keypress-over-socket mode
    if (argc == 1)
    {
        socket_server_loop();
   }

        // If filename passed, then 
        // read sin_table and writedata
        if (argc == 2){
            char *sin_file= argv[1];
            load_wave_table(sin_file);
            write_sample_data();
        }

        if (argc == 3){
    vga_ball_color_t c;
        char *a_val = argv[1];
        char *r_val = argv[2];
       short attack = (short) strtol(a_val, NULL, 10);
       short release = (short) strtol(r_val, NULL, 10);
       printf("\nWriting Attack:%d\tRelease:%d\n", attack, release);
       
       typedef union {
            uint16_t val;
            struct{
                unsigned int att: 8;
                unsigned int rel: 8;
            };
        }packed_un_t2;

        packed_un_t2 u2 = {.val = 0};
        u2.att = attack;
        u2.rel = release;
        c.sin = u2.val;
        c.loc = ENV_WRITEADDR;
        set_background_color(&c);
 
        }


    printf("VGA BALL Userspace program terminating\n");
    return 0;
}
